
# agohGiuga.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# https://en.wikipedia.org/wiki/Agoh%E2%80%93Giuga_conjecture
# https://planetmath.org/agohgiugaconjecture
'''
The Agoh-Giuga prime conjecture states that a number p is prime if and only if it satisfies the congruence:
1^(p−1)+2^(p−1)+⋯+(p−1)^(p−1) ≡ −1 (modp)
'''

from fractions import Fraction as F
import math
import bernoulli


def agoh_giugaB(p):
  if p <= 1:
    return False  
  bern = p * bernoulli_number(p-1)
  fracMod = fracModulo(bern.numerator, bern.denominator, p)
  return fracMod == p-1


def bernoulli_number(n):
  # https://en.wikipedia.org/wiki/Bernoulli_number
  A = [0] * (n+1)
  for m in range(n+1):
    A[m] = F(1, m+1)
    for j in range(m, 0, -1):
      A[j-1] = j*(A[j-1] - A[j])
  return A[0] # (which is Bn)


def fracModulo(a, b, m):
  # return (a/b) % m
  # https://www.geeksforgeeks.org/expressing-a-fraction-as-a-natural-number-under-modulo-m/
  c = math.gcd(a, b)
  a = a // c
  b = b // c
  # (b ^ m-2) % m
  d = pow(b, m-2, m)
  return ((a % m) * (d % m)) % m


# ---------------------------------------

# check the Agoh-Giuga prime conjecture for a number p
def check_agoh_giuga(p):
  if p <= 1:
    return False 
  tot = sum(pow(i, p-1, p) for i in range(1, p))
  return tot%p == p-1



limit = 100

print("\nCheck 2 to", limit, "using Bernoulli Agoh-Giuga")
for p in range(2, limit + 1):
  if agoh_giugaB(p):
    print(p, end=' ')
print()

print("\nCheck 2 to", limit, "using Agoh-Giuga")
for p in range(2, limit + 1):
  if check_agoh_giuga(p):
    print(p, end=' ')
print()


