
# carmichael.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": Carmichael numbers
# https://en.wikipedia.org/wiki/Carmichael_number
# Also see pseudoPrimes.py
'''
A Carmichael number is a composite number n  which 
in modular arithmetic satisfies the congruence relation:
    b^n ≡ b (mod n)
for all integers b. The relation may also be expressed 
in the form:
    b^(n − 1) ≡ 1 (mod n) 
for all integers b that are relatively prime to n. 
Also see fermatLittle.py
'''

import pUtils
import math


def isCarmichael1(n):
  b = 2
  while b < n :
    if math.gcd(b, n) == 1:
      if pow(b, n-1, mod=n) != 1:
        return False
    b += 1
  return True


def isCarmichael2(n):
  if pUtils.isPrime(n):
    return False
  factors = pUtils.primeFactors(n)
  # Korselt's criterion: n must be square-free
  if any(n % (p**2) == 0 for p in factors):
    return False
  # For all prime divisors p, (p-1) must divide (n-1)
  for p in factors:
    if (n-1) % (p-1) != 0:
      return False
  return True


def carmichaels(limit):
  cNums = []
  for n in range(3, limit, 2):
    # if isCarmichael1(n):    # slow
    if isCarmichael2(n):      # faster
      cNums.append(n)
  return cNums


limit = 10000
cNums = carmichaels(limit)
print(f"Carmichael numbers up to {limit}:")
for cn in cNums:
  print(" ", cn, pUtils.factorize(cn))

