
# chinese.py
# https://en.wikipedia.org/wiki/Chinese_remainder_theorem
# https://en.wikipedia.org/wiki/Chinese_remainder_theorem
''' An excellent video:
     "The Chinese Remainder Theorem made easy"
     Randell Heyman
     https://www.youtube.com/watch?v=ru7mWZJlRQg

  The Chinese Remainder Theorem (CRT) allows you to solve 
  system of congruences using modular arithmetic. 

  Return the smallest x such that: 
    x ≡ r1 mod m1
    x ≡ r2 mod m2
       :
    x ≡ rn mod mn

  The CRT guarantees a unique solution for x
  modulo M=m1×m2×...×mn.

  You can solve it using Extended GCD to find the 
  modular inverses needed for the solution.

  --------------
  Apply CRT to speed up the multiplication of large numbers:

  1. Select several relatively small pairwise coprime moduli (e.g., prime numbers).

  2. Represent each of the large numbers as their remainders when divided by the selected moduli.

  3. Perform modular multiplications

  4. Reconstruct the result: Use CRT to combine the results and get the final product modulo the product of all moduli.
'''

import math
import pUtils


def crt(mods, rems):
  # find x (mod m1*m2*...*mn)
  modProd = math.prod(mods)
  x = 0
  for m, r in zip(mods, rems):
    partP = modProd // m   # exclude current m
    # Add the contribution from this congruence
    x += partP * pUtils.modInverse(partP, m) * r 
         # first two parts gets 1 (mod m)
  return x % modProd



def fastMultiply(a, b, moduli):
  # Decompose the large numbers into remainders
  a_rems = [a%m for m in moduli]
  b_rems = [b%m for m in moduli]
  print("a_rems:", a_rems)
  print("b_rems:", b_rems)

  # Multiply the remainders
  multRems = [(a_rem * b_rem)%m 
       for a_rem, b_rem, m in zip(a_rems, b_rems, moduli)]
  print("multiplied remainders:", multRems)

  # Use CRT to reconstruct the result
  return crt(moduli, multRems)


# Sunzi example
rems = [2, 3, 2]  # Remainders
mods = [3, 5, 7]  # Moduli  (pairwise coprime)
print("Rems =", rems, " Mods =", mods)
print("x =", crt(mods, rems))  # 23

rems = [2, 2, 1] 
mods = [3, 4, 5] 
print("Rems =", rems, " Mods =", mods)
print("x =", crt(mods, rems))  # 26

# -----------------------
moduli = [3, 5, 7, 11, 13]  
    # Moduli must be pairwise coprime
# large numbers to multiply
a = 12345678901234567890
b = 98765432109876543210
print(a, "*", b, "mod", math.prod(moduli))
res = fastMultiply(a, b, moduli)
print("Fast multiplication using CRT:")
print(" ", res)  # 1200
