
# conway.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": Conway's prime-producing machine
# https://en.wikipedia.org/wiki/FRACTRAN
# https://oeis.org/A007542
'''
Also known as PRIMEGAME.

Other examples:

n = 6
fracs = [(228, 131), (327, 158), (161, 394), (77, 425)]
iters = 4

n = 2
fracs = [(345, 162), (108, 125), (90, 45), (293, 277)]
iters = 5
'''

import math

def fractran(n, fracs, iters):
  result = [n]
  numerator = n
  denominator = 1
  for _ in range(iters):
    for j in range(len(fracs)):
      numer = fracs[j][0] * numerator
      denom = fracs[j][1] * denominator
      div = math.gcd(numer, denom)
      numer //= div
      denom //= div
      if denom == 1:
        result.append(numer)
        numerator = numer
        denominator = denom
        break
  return result


def isPower2(n):
  ''' powers of 2 have exactly one 1 bit in their binary 
    representation, and subtracting 1 flips all the lower 
    bits, making the bitwise AND zero.
  '''
  return n > 0 and (n & (n - 1)) == 0

# ------ main ----------


n = 2
fracs = [(17,91), (78,85), (19,51), (23,38),
         (29,33), (77,29), (95,23), (77,19),
         ( 1,17), (11,13), (13,11), (15, 2),
         ( 1, 7), (55, 1) ]
iters = 100000
print("n, iters, fracs:", n, iters, fracs)

fracSeq = fractran(n, fracs, iters)
pows2 = [(i,f) for (i,f) in enumerate(fracSeq) if isPower2(f)]
print("\n(Pos, Power):", pows2)
print("\nLog Primes:", [int(math.log2(p)) for (_,p) in pows2])
