
# diffieHelman.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# https://en.wikipedia.org/wiki/Diffie%E2%80%93Hellman_key_exchange
# Also see primitiveRoots.py, rsa.py
'''
Alice and Bob agree on a prime and a 
generator. This allows Alice and Bob to 
establish a mutual key (mod prime), with neither able 
to infer the other's secret key. 
'''

import random
import pUtils


def primitiveRoot(prim):
  # find the smallest primitive root modulo prim
  # https://en.wikipedia.org/wiki/Primitive_root_modulo_n
  if prim == 2:
    return 1
  facs = pUtils.primeFactors(prim-1)  
  for root in range(2, prim):
    for f in facs:
      if pow(root, (prim-1)//f, prim) == 1:
        break
    else:
      return root
  print("Error in primitiveRoot()")
  return None  # Should never reach here


def genKeys(gen, prim):
  privKey = random.randint(2, prim-2) 
  pubKey = pow(gen, privKey, prim)
  return privKey, pubKey


# -------------------------

prim = pUtils.prime( random.randint(100, 500))  
gen = primitiveRoot(prim)

print("Alice's (private key, public key):")
alice_priv, alice_pub = genKeys(gen, prim)
print((alice_priv, alice_pub))

print("Bob's (private key, public key):")
bob_priv, bob_pub = genKeys(gen, prim)
print((bob_priv, bob_pub))

print("Bob's mutual key (using Alice's public key):")
bob_mutual_key = pow(alice_pub, bob_priv, prim)
print(bob_mutual_key)

print("Alice's mutual key (using Bob's public key):")
alice_mutual_key = pow(bob_pub, alice_priv, prim)
print(alice_mutual_key)

# Check that both mutual keys are the same
#  -- a shared secret key
print("Mutual keys match?", alice_mutual_key == bob_mutual_key)

