
# drawPrimeSums.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# inputs: 59 or 139
'''
  A circle of odd integers 3 to n are drawn (along with 2)
  and each integer is split into a sum of unique primes. 

  Arrows are drawn to those primes, which are arbitrarily 
  colored to make them easier to see.

  The circles and lines are drawn with Graphviz
  (pip install graphviz)
  https://graphviz.readthedocs.io/en/stable/manual.html

  Graphviz is also used in modmult.py
'''

import math
import pUtils
import graphviz

RADIUS = 2

def positionNodes(g, nums):
  g.node_attr['shape']='circle'
  g.node_attr['fontsize']='6'
  g.node_attr['width']='0.2'
  g.node_attr['height']='0.2'
  g.node_attr['fixedsize']='true'
  g.node_attr['penwidth']='0.5'

  angleIncr = math.pi*2/(len(nums)+1)  # for '2'
  g.node('2', pos=str(RADIUS)+",0!")
  i = 1
  for n in nums:
    x = RADIUS*math.cos(angleIncr*i)
    y = RADIUS*math.sin(angleIncr*i)
    g.node(str(n), pos=str(x)+","+str(y)+"!")
    i += 1


def primeSum(n):
  # return a list holding a prime sum that == n
  if n % 2 == 0:
    print("The number must be odd.")
    return []
  prims = pUtils.primes(n)
  return pSum(n, prims, [])


def pSum(n, prims, partial):
  tot = sum(partial)
  if tot == n:
    return partial
  if tot > n:
    return []
  for i, p in enumerate(prims):
    restPrims = prims[i+1:]
    res = pSum(n, restPrims, partial+[p])
    if res != []:
      return res
  return []


def getCol(src):
  ''' Lines are arbitrarily colored to make them easier to
    distinguish. The color is based on if the decimal
    remainder of the source number appears below
  '''
  rem = src%10
  if rem == 1:
    return 'red'
  elif rem == 3:
    return 'darkgreen'
  elif rem == 5:
    return 'blue'
  elif rem == 7:
    return 'brown'
  elif rem == 9:
    return 'gold4'
  else:
    return 'black'



limit = int(input("odd n? ")) 

# print sum for 17 without using Graphviz
print(17, "=", primeSum(17))

g = graphviz.Digraph('G', engine='neato', 
                 filename='drawPrimeSums.gv')
g.graph_attr['label']="Odd ints 3-" + str(limit) + \
                            " as unique prime sums"
g.graph_attr['bgcolor'] ='grey72'

g.edge_attr['penwidth']='0.3'
g.edge_attr['arrowsize']='0.5'

nums = list(range(3, limit+1, 2))

positionNodes(g, nums)  

for n in nums:
  ps = primeSum(n)
  nStr = str(n)
  for p in ps:
    g.edge(nStr, str(p), color=getCol(n)) # n --> p

g.view()
