
# eulerTotOthers.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": Euler, Leonhard (1707-1783), Euler’s totient (phi) function
# https://en.wikipedia.org/wiki/Euler%27s_totient_function
'''
Euler’s totient (phi) function: the number of 
integers less than n and coprime to it. 

Other approaches from:
  "Intro to number theory lecture. 
  14. Euler's totient function"
  Richard E Borcherds
https://www.youtube.com/watch?v=8I0z_Lobtso&list=PL8yHsr3EFj53L8sMbzIhhXSAOpuZ1Fov8&index=14

Also see eulerTotient.py
'''

import math
import itertools
import pUtils

def totientCR(n):
  return len( coprimeResidues(n))


def coprimeResidues(n):
  # all residue classes modulo n that are coprime to n
  if n == 1:
    return [1]
  else:
    return [i for i in range(1, n) if math.gcd(i, n) == 1]


def totientPF(n):
  # Totient using prime factors
  if n == 1:
    return 1
  factors = pUtils.primeFactors(n)
  result = n
  for p in factors:
    result *= (1 - 1/p)
  return int(result)


def totientIE(n):
  # Totient using inclusion-exclusion
  # ϕ(n)= n − |A1|−|A2|− ... + |A1,2| + ...
  if n == 1:
    return 1
  factors = pUtils.primeFactors(n)
  tot = n 
  nFacts = len(factors)
  for k in range(1, nFacts+1):
    # Iterate over all combinations of size k
    for subset in itertools.combinations(factors, k):
      prod = math.prod([f for f in subset])
      # alternate between subtracting and adding as k changes
      if k % 2 == 1:
        tot -= n//prod
      else:
        tot += n//prod
  return tot


print("pUtils phi():")
print(" ", *[pUtils.phi(i) for i in range(1, 21)])
print()

print("phi() using prime factors:")
print(" ", *[totientPF(i) for i in range(1, 21)])
print()

print("phi() using inclusion-exclusion:")
print(" ", *[totientIE(i) for i in range(1, 21)])
print()

print("phi() using coprime residues:")
for i in range(1, 21):
  print(i, ".", coprimeResidues(i), "len:", totientCR(i))
