
# eulerTotient.py 
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": Euler, Leonhard (1707-1783), Euler’s totient (phi) function
# https://en.wikipedia.org/wiki/Euler%27s_totient_function
'''
  Euler’s totient (phi) function: the number of integers 
  less than n and  coprime to it, denoted by phi(n). 
  Its first few values are:  
    n     : 1 2 3 4 S 6 7 8 9 10 11 12 
    phi(n): 1 1 2 2 4 2 6 4 6 4  10 4 
  If p is prime, phi(p) = p-1, and phi(p^a) = p^(a-1)(p-1)
  e.g. 
  phi(25) = phi(5^2) = 5^1 * (5-1)  = 5*4 = 20
  phi(27) = phi(3^3) = 3^2 * (3-1)  = 9*2 = 18

  Also see eulerTotPlot, eulerTotOthers.py
'''

import math, random
import pUtils

def randomInts(start, end, count):
  # return a sorted list of count number of integers, 
  # randomly selected from the range start--end
  ns = random.sample(range(start, end+1), count)
  return sorted(ns)


# generate a table of phi values from 1 to limit
limit = 100
print("  n  phi(n)  ['*' means n is prime]")
COLS = 5
for r in range(1, limit+1, COLS):
  for x in range(r, r+COLS):
    print(f"{x:3d} {pUtils.phi(x):4d}", 
          "*" if pUtils.isPrime(x) else " ", end ="  | ")
  print()

# ----------------------------------
'''
phi(n)/n < 1 for every integer n > 1. As n grows very
large, the average value of phi(n)/n approaches 6/pi^2. 
Also see eulerTotNNPlot.py.
'''

print("\nEstimate an average for phi(n)/n using 500 random n's:")
xs = randomInts(2, 10000, 500)
ys = [ pUtils.phi(x)/x for x in xs ]
print(f"Average: {(sum(ys) / len(ys)):.4f}")
print(f"Compare to 6/pi^2 = {6/(math.pi**2):.4f}")



