
# fermatLittle.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": Fermat's Little Theorem
'''
Fermat's Little Theorem states that if p is a prime 
and a and p are coprime by p, then:
a^(p−1) ≡ 1 (mod p)
or
a^p ≡ a (mod p)

e.g.
2^3 ≡ 2 (mod 3)
Fermat's Little Theorem is not a definitive primality 
test, as certain composite numbers (Carmichael numbers) 
can also satisfy the theorem.
Also see carmichael.py

Video:
"Introduction to Number Theory (Berkeley Math 115)
10. Applications of Fermat's Theorem"
Richard E Borcherds

https://www.youtube.com/watch?v=fgHEKAdErbU&list=PL8yHsr3EFj53L8sMbzIhhXSAOpuZ1Fov8&index=10

'''
# inputs:  91  or  97

import math, random


def isFermat(p, a):
  if p <= 1:
    return False  
  if a % p == 0:
    return False 
  # Fermat's little theorem: a^(p-1) ≡ 1 (mod p)
  return pow(a, p-1, p) == 1


def isProbPrime(p, k=5):
  if p == 2:  # 2 is prime
    return True
  if p <= 1 or p % 2 == 0:  
    return False
  
  # Perform the test k times with random values
  for _ in range(k):
    a = random.randint(2, p-1) 
    if not isFermat(p, a):
      return False
  return True  # probably prime


def orderMod(a, p):
  # compute the order of an a modulo p;
  # find the smallest k such that a^k ≡ 1 (mod p)
  if math.gcd(a, p) != 1:
    return None  # No order exists
  k = 1
  curr = a % p 
  while curr != 1:
    curr = (curr * a) % p
    k += 1
  return k


p = int(input("p? "))
print(f"{p} (probably) prime? {isProbPrime(p)}")

# ---------------
print("\nOrders for 1-6 for p=7")
for i in range(1, 7):  # order a mod p divides p-1
  print(" ", i, orderMod(i, 7))
