
# gammaL.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# https://en.wikipedia.org/wiki/Gamma_function
# https://en.wikipedia.org/wiki/Lanczos_approximation
# Also see gammaPlot.py

import math, cmath
from frange import linspace
import matplotlib.pyplot as plt

# Lanczos approximation constants
g = 7
p = [
  0.99999999999980993,
  676.5203681218851,
  -1259.1392167224028,
  771.32342877765313,
  -176.61502916214059,
  12.507343278686905,
  -0.13857109526572012,
  9.9843695780195716e-6,
  1.5056327351493116e-7
]


def gammaL(z):
  if z.real < 0.5:
    # Reflection formula: 
    # Γ(z) * Γ(1 - z) = π / sin(πz)
    return math.pi / (cmath.sin(math.pi*z) * gammaL(1-z))
  else:
    z -= 1
    bigA = p[0]
    for i in range(1, g+2):
      bigA += p[i] /(z+i)
    t = z + g + 0.5
    return math.sqrt(2*math.pi) * \
               t**(z+0.5) * cmath.exp(-t)*bigA



if __name__ == "__main__":
  print("gammaL(-5):", gammaL(-5))
  print("gammaL(5):", gammaL(5))  # 4!
  print("gammaL(0.5):", gammaL(0.5)) # sqrt(pi)
  print("gammaL(-0.5):", gammaL(-0.5))
  print("gammaL(-1.5):", gammaL(-1.5))
  print()
  print("gammaL(3 + 2j):", gammaL(3 + 2j))   # Complex
  print("gammaL(-2.5 + 1j):", gammaL(-2.5 + 1j))  
  print("gammaL(0.9 + 14j):", gammaL(0.9 + 14j)) 
  print()

  xs = linspace(-5, 5, 500)
  lgs = [gammaL(x).real for x in xs]

  plt.plot(xs, lgs, color='red')

  plt.axhline(y=0, color='k')
  plt.axvline(x=0, color='k')
  plt.title("Lanczos Approximation for $\Gamma(x)$")
  plt.xlabel("x")
  plt.ylabel("$\Gamma(x)$")
  plt.ylim(-4, 4)
  plt.grid(True)
  plt.show()