
# gaussCircle.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": Gauss's circle problem
'''
https://mathworld.wolfram.com/GausssCircleProblem.html
https://en.wikipedia.org/wiki/Gauss_circle_problem
https://oeis.org/A000328

See the video:
 "Pi hiding in prime regularities", 3Blue1Brown
 https://www.youtube.com/watch?v=NaL_Cb42WyY
   - but without using complex numbers
'''
import math
import matplotlib.pyplot as plt
from matplotlib.patches import Circle


def lattice(r):
  return 1 + 4*r + 4*sum([int(math.sqrt(r**2 - i**2))
                               for i in range(1, r+1)]) 

def genPoints(r):
  # all integers (m,n) such that m^2 + n^2 <= r^2
  pairs = []
  for m in range(-r, r+1):
    for n in range(-r, r+1):
      if m**2 + n**2 <= r**2:
        pairs.append((m, n))
  return pairs



latlist = [lattice(r) for r in range(10)]
print("First Lattice counts:  ", latlist)
ptslist = [len(genPoints(r)) for r in range(10)]
print("First genPoints counts:", ptslist)

fig, ax = plt.subplots(1, 2, figsize=(10, 5))  
                  # no. ros, cols

# -------- visualise pi approx ---------

xs = list(range(1, 301))
ys = [ lattice(x)/(x*x) for x in xs]

ax[0].plot(xs, ys)
ax[0].set_xlabel("x")
ax[0].set_ylabel("lattice(x)/x^2")
ax[0].set_ylim(3.13, 3.15)
ax[0].axhline(y=math.pi, color='red')
ax[0].text(150, math.pi+0.0005, "y=$\pi$", color='red')
ax[0].set_title('Pi Approximations')


# ------- visualise a small lattice --------
limit = 5
xs, ys = zip(*genPoints(limit))

ax[1].scatter(xs, ys, color='black', s=40, marker=',')
circle = Circle((0,0), limit, lw=2, color='red', fill=False)
ax[1].add_patch(circle)
ax[1].set_aspect('equal')
ax[1].set_title(f'Lattice Points in (-{limit}, {limit})')
ax[1].axis('off')

plt.show()
