
# mersenne.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": Mersenne numbers and Mersenne primes
# https://en.wikipedia.org/wiki/Mersenne_conjectures
# https://en.wikipedia.org/wiki/Gillies%27_conjecture
# https://t5k.org/mersenne/heuristic.html
# Also see lenstra.py, lucasLehmer.py
'''
Mersenne Primes are primes of the form M_n = 2^n−1, where 
n is a positive integer. The Mersenne Prime Conjecture 
suggests that for certain values of n, M_n is prime.

While not all numbers of this form are prime, there are 
known prime values for specific n, including  
n = 2,3,5,7,13,17,19,31,61,89,107,127
'''

import math
import pUtils
import matplotlib.pyplot as plt
import numpy as np   # used for line-fitting


def mersennePrimes(limit):
  mprims = []
  for n in range(2, limit+1):
    val = (1 << n) - 1  # Equivalent to 2**n - 1
    if pUtils.isPrimeMR(val):  # was isPrime()
      mprims.append((n, val))
  return mprims


# hunting for Mersenne primes
limit = 610   # 2300  # 610  # 128  
mprims = mersennePrimes(limit)
print(f"Mersenne primes for n up to {limit}:")
idx = 1
for n, mp in mprims:
  print(f"{idx:2d}. M_{n} = 2^{n} - 1 = {mp}")
  idx += 1

print("\n", len(mprims), "n's in M_n:", [n for (n,mp) in mprims])


# ---------------------------------------
# how many Mersenne primes?

mpLast = mprims[-1][1]
mpnLast = mprims[-1][0]
print("\nGilles' Estimated no of primes < M_", mpnLast, "=",
       f"{(2/math.log(2)) * math.log( math.log(mpLast)):.3f}" )
print(f"Constant: {(2/math.log(2)):.4f}")

C = math.exp(pUtils.GAMMA) /math.log(2)
print(f"\nLenstra–Pomerance–Wagstaff Estimated no of primes < M_{ mpnLast} =",
       f"{C * math.log( math.log(mpLast)):.3f}" )
print(f"Constant: {C:.4f}")


xs = [math.log(math.log(mp)) for (n,mp) in mprims]
ys = list(range(1, len(mprims)+1))

plt.scatter(xs, ys, label='Data points')

# fit a straight line y = mx + c
m, c = np.polyfit(xs, ys, 1)

# Plot points for the fitted line
yfits = [ (m*x)+c for x in xs]
plt.plot(xs, yfits, color='red', 
       label=f'Fitted line: y = {m:.4f}x + {c:.4f}')

plt.xlabel('log(log($M_n$))')
plt.ylabel('No. of Mersenne primes')
plt.title('log(log($M_n$)) and fitted line')
plt.legend()
plt.show()
''' There's a more detailed graph at
  http://m-brella.be/math/topics/MersennePrime.html
'''


