
# mertensConst.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": Mertens constant
# https://en.wikipedia.org/wiki/Meissel%E2%80%93Mertens_constant
'''
The sum of the reciprocals of the primes diverges, 
but very slowly. Let P(p)=1/2 + 1/3 + 1/5 + 1/7 + ... + 1/p
The sum from 1/2 to 1/p is approximately log log p, and 
we have the result that P(p) — log log p tends 
to the Mertens constant, 0.261497212847642... as p goes 
to infinity.
'''

import math
import pUtils
import matplotlib.pyplot as plt

def harmonic_sum_of_primes(n):
  prims = pUtils.primes(n+1)
  return sum([1/p for p in prims])


def mertens(n):
  return harmonic_sum_of_primes(n) - math.log(math.log(n))


n = 10**6 
print(f"Approx of Mertens' constant with n = {n}: {mertens(n)}")

# n from 100 to 1,000,000
ns = [10**i for i in range(2, 7)]  
hs = [harmonic_sum_of_primes(n) for n in ns]
llVals = [math.log(math.log(n)) for n in ns]

plt.plot(ns, hs, label="Sum of Reciprocals of Primes < n")
plt.plot(ns, llVals, label=r"$\ln(\ln(n))$")

# arrow indicating the Mertens constant separation
x =10**4
y1 = harmonic_sum_of_primes(x)
y2 = math.log(math.log(x))
plt.annotate('', xy=(x,y1), xytext=(x,y2),
    arrowprops=dict(arrowstyle='<->', color='red', lw=2))
plt.text(x, (y1+y2)/2, ' M = 0.26145...', color='k', 
    ha='left', va='center')

plt.xscale('log')
plt.title("Sum of Reciprocals of Primes < n vs ln(ln(n))")
plt.xlabel("n (log scale)")
plt.ylabel("Values")
plt.legend()
plt.grid(True)
plt.show()




