
# mobius.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": randomness, of primes, Von Sternach and a prime random walk
# https://en.wikipedia.org/wiki/M%C3%B6bius_function
'''
The Möbius function, μ(n):

 * μ(n) = 1 if n is a square-free positive integer with 
   an even number of prime factors;

 * μ(n) = −1 if n is a square-free positive integer with 
   an odd number of prime factors;

 * μ(n) = 0 if n has a squared prime factor.
'''

import matplotlib.pyplot as plt
import math
import pUtils


def mobius(n):
  fs = pUtils.primeFactors(n)
  if any(n%(p*p) == 0 for p in fs):
    return 0  # n is divisible by a square prime
  return -1 if (len(fs)%2 == 1) else 1


# Von Sternach random walk
limit = 100001
mobZeros = len([n for n in range(2, limit+1)
                              if (mobius(n) == 0)])
r = limit-1
print(f"Prob of non-zero Mobius for 2-{limit}: {(r-mobZeros)/r}")
print(f"Compare with 6/pi^2: {(6/(math.pi**2)):.5f}")

# ------------------------------------------
ns = range(1, 101)
mobs = [mobius(n) for n in ns]

plt.figure(figsize=(10, 4))
plt.scatter(ns, mobs, s=20)
plt.axhline(y=0, color='k')
plt.axvline(x=0, color='k')
plt.title("Möbius Function $\mu(n)$")
plt.xlabel("$n$")
plt.ylabel("$\mu(n)$")
plt.ylim(-1.2, 1.2)
plt.show()
