
# pseudoStrong.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": pseudoprimes, strong
# https://en.wikipedia.org/wiki/Strong_pseudoprime
'''
A composite number n is called a strong pseudoprime to 
the base a if it satisfies the conditions of the 
Miller-Rabin test, meaning that for n−1 = 2^s*d where 
d is odd, one of the following holds:

1. a^d ≡ 1 (mod n)
2. a^(2^r*d) ≡ −1 (mod n) for some 0 ≤ r < s
'''

import random
import pUtils


def isStrongPP(n, a):
  # Miller-Rabin test for strong pseudoprimes
  if n < 2 or n % 2 == 0:
    return False
  if pUtils.isPrime(n):
    return False 
  
  # Write n-1 as 2^s*d where d is odd
  d = n - 1
  s = 0
  while d % 2 == 0:
    d //= 2
    s += 1
  # Test condition a^d ≡ 1 (mod n)
  if pow(a, d, n) == 1:
    return True
  # Test condition a^(2^r * d) ≡ -1 (mod n) for some r < s
  for r in range(s):
    if pow(a, 2**r*d, n) == n-1:
      return True
  return False


def gen_strong_pps(limit, a):
  return [n for n in range(3, limit+1, 2)
                            if isStrongPP(n, a)]


def testBase(a, limit):
  print(f"Strong pseudoprimes to base {a} up to {limit}:")
  print(gen_strong_pps(limit, a))


limit = 10000
testBase(2, limit)
testBase(3, limit)
testBase(5, limit)
