
# riemannZeta.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": Riemann hypothesis
# https://en.wikipedia.org/wiki/Riemann_zeta_function
# https://mathworld.wolfram.com/RiemannZetaFunction.html
'''
  Tests of three variants of the Riemann Zeta function
  which can be compared to 'correct' values at https://solvemymath.com/online_math_calculator/number_theory/riemann_function/index.php
    -- zetac() is the winner

  The magnitudes of zetac(complex(0.5, t)) close
  to 0 are printed for -30 <= t <= 30 and all of 
  the magnitudes in this range are plotted.

  Also see riemannZeta2.py and riemannZeta3.py
'''

import matplotlib.pyplot as plt
import math
import itertools


def zeta(s, numTerms=1000):
  # only converges when Re(s) > 1
  return sum([1.0/(n**s) for n in range(1, numTerms+1)])


def zetaD(s, numTerms=1000):
  # using the Dirichlet eta function
  # (https://en.wikipedia.org/wiki/Dirichlet_eta_function)
  # large numerical errors
  return sum([(-1**(n+1))/(n**s) for n in range(1, numTerms+1)]) / (1 - 2**(1-s))


def zetac(s, numTerms=100):
  # using the Dirichlet eta function
  # (https://en.wikipedia.org/wiki/Dirichlet_eta_function)
  # which is convergent when the real part > 0 but is
  # Abel summable for any complex number;
  # see the Numerical algorithms subsection on the Wiki page
  if s == 1: 
    return complex("inf")
  term = ( 1/2**(n+1) * \
      sum((-1)**k * math.comb(n,k) * (k+1)**(-s) 
                 for k in range(n+1)) 
                    for n in itertools.count(0))
  return sum(itertools.islice(term, numTerms))/(1 - 2**(1-s))



print("zeta(2)", zeta(2), math.pi**2/6)
print("zeta(4)", zeta(4), math.pi**4/90)
print("zeta(2+i)", zeta(2+1j))
print("zeta(0.5)", zeta(0.5))
print("zeta(-1)", zeta(-1), -1/12)
print("zeta(-2)", zeta(-2))
print()

print("zetaD(2)", zetaD(2), math.pi**2/6)
print("zetaD(4)", zetaD(4), math.pi**4/90)
print("zetaD(2+i)", zetaD(2+1j))
print("zetaD(0.5)", zetaD(0.5))
print("zetaD(-1)", zetaD(-1), -1/12)
print("zetaD(-2)", zetaD(-2))
print()

print("zetac(2)", zetac(2), math.pi**2/6)
print("zetac(4)", zetac(4), math.pi**4/90)
print("zetac(2+i)", zetac(2+1j))
print("zetac(0.5)", zetac(0.5))
print("zetac(-1)", zetac(-1), -1/12)
print("zetac(-2)", zetac(-2))
print()

# -----------------------------------------

print("t values where | zetac(complex(0.5, t)) | < 0.1:")
ts = []
mags = []
t = -30.0
while t <= 30.0:
  z = zetac(complex(0.5, t))
  mag = math.sqrt(z.real**2 + z.imag**2)
  if mag < 0.1:
    print(f"  {t:>8.4f}")
  ts.append(t)
  mags.append(mag)
  t += 0.2


plt.figure(figsize=(10, 6))
plt.plot(ts, mags)
ax = plt.gca()
ax.axhline(y=0, color='k')
ax.axvline(x=0, color='k')
plt.title(r'Magnitude of the Riemann Zeta Function $\vert\zeta(1/2 + it)\vert$')
plt.xlabel('t')
plt.xticks(range(-30, 31, 2))
plt.ylabel(r'$\vert\zeta(1/2 + it)\vert$')
plt.grid()
plt.show()
