
# riemannZeta2.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": Riemann hypothesis
# https://en.wikipedia.org/wiki/Riemann_zeta_function
# https://mathworld.wolfram.com/RiemannZetaFunction.html
'''
  Plots the polar graph of zetac(complex(0.5, t)) for 
  0 <= t < 35. The first five zeros in the critical strip 
  occur where the spiral passes through the origin.

  Also see:
    "But what is the Riemann zeta function? 
     Visualizing analytic continuation", 3Blue1Brown
     https://www.youtube.com/watch?v=sD0NjbwqlYw

  Also see riemannZeta1.py and riemannZeta3.py
'''

import matplotlib.pyplot as plt
import math
import itertools


def zetac(s, numTerms=100):
  # using the Dirichlet eta function
  # (https://en.wikipedia.org/wiki/Dirichlet_eta_function)
  # which is convergent when the real part > 0 but is
  # Abel summable for any complex number;
  # see the Numerical algorithms subsection
  if s == 1: 
    return complex("inf")
  term = ( 1/2**(n+1) * \
      sum((-1)**k * math.comb(n,k) * (k+1)**(-s) 
                 for k in range(n+1)) 
                    for n in itertools.count(0))
  return sum(itertools.islice(term, numTerms))/(1 - 2**(1-s))


reals = []; imags = []
t = 0.0
while t <= 35.0:
    z = zetac(complex(0.5, t))
    reals.append(z.real)
    imags.append(z.imag)
    t += 0.05


plt.plot(reals, imags)
ax = plt.gca()
ax.axhline(y=0, color='k')
ax.axvline(x=0, color='k')
ax.axis('equal')
plt.title(r'Polar graph of $\zeta(1/2 + it)$, t=0 to 35')
plt.xlabel('Re $\zeta(1/2 + it)$')
plt.ylabel(r'Im $\zeta(1/2 + it)$')
plt.show()
