
# riemannZeta3.py
# Andrew Davison, ad@coe.psu.ac.th, Nov. 2024
# David Wells - "Prime Numbers": Riemann hypothesis
# https://en.wikipedia.org/wiki/Riemann_zeta_function
# https://mathworld.wolfram.com/RiemannZetaFunction.html
'''
  Plots a contour map of the phase angle of zetac() 
  versus the real part ReS of the complex variable s 
  (the x-axis) and the imaginary part ImS of the 
  complex variable s (the y-axis) in the range -20 to 20.

  The colorbar shows the values of the 
  phase angle in the interval -180 to 180.

  The phase shifts along the "ribs" of the zeta function. 
  Along the critical line, every time the function changes 
  sign, a discontinuous jump by pi occurs in the phase.

  WARNING: this code takes about 60-70 secs to run
  (other timings listed below)!!

  Also see riemannZeta2.py and riemannZeta3.py
'''


import matplotlib.pyplot as plt
import math, cmath
import itertools, time
from frange import frange


def zetac(s, numTerms=100):
  # using the Dirichlet eta function
  # (https://en.wikipedia.org/wiki/Dirichlet_eta_function)
  # which is convergent when the real part > 0 but is
  # Abel summable for any complex number;
  # see the Numerical algorithms subsection
  if s == 1: 
    return complex("inf")
  term = ( 1/2**(n+1) * \
      sum((-1)**k * math.comb(n,k) * (k+1)**(-s) 
                 for k in range(n+1)) 
                    for n in itertools.count(0))
  return sum(itertools.islice(term, numTerms))/(1 - 2**(1-s))


start = time.time()
phis = []
coords = list(frange(-20, 20, 0.5))
for imS in coords:
  row = []
  for reS in coords:
    _, phi = cmath.polar(zetac(complex(reS, imS)))
    row.append(math.degrees(phi))
  phis.append(row)
print(f"Elapsed time: {(time.time() - start):.2f} secs")
    # 0.5 step takes 69 secs
    # 0.1 step takes 1700 secs
    # 0.05 step takes 8097.54 secs

hm=plt.imshow(phis, cmap='rainbow', 
                       extent=[-20, 20, -20, 20], 
                       vmin=-180, vmax=180)
cbar = plt.colorbar(hm) 
cbar.set_ticks(range(-180, 181, 60))
cbar.set_label('degrees')

plt.xlabel('ReS')
plt.ylabel('ImS')
plt.title("Contour map of the phase of the zeta function")
plt.show()